/*$Id: d_mosmod.cc,v 12.1 1998/10/16 20:53:15 al Exp $ -*- C++ -*-
 * mos model basics
 * netlist syntax:
 * device:  mxxxx d g s b mname <device args> <model card args>
 * model:   .model mname NMOS <args>
 *	or  .model mname PMOS <args>
 */
#include "ap.h"
#include "d_diode.h"
#include "d_mos.h"
#include "error.h"
#include "io.h"
#include "l_lib.h"
#include "u_opt.h"
/*--------------------------------------------------------------------------*/
//		MOS_COMMON::MOS_COMMON();
//		MODEL_MOS::MODEL_MOS(const char*);
//	void	MODEL_MOS::parse(CS& cmd);
// 	void	MODEL_MOS::print(int,int)const;
//	void	MODEL_MOS::tr_eval(CARD*)const;
/*--------------------------------------------------------------------------*/
const double DEFAULT_vto   = 0.0;		/* last resort defaults */
const double DEFAULT_gamma = 0.0;		/* in case can't calculate */
const double DEFAULT_phi   = 0.6;
const double DEFAULT_cj    = 0.0;
const double DEFAULT_tox   = 1e-7;
const double DEFAULT_nss   = 0.0;
const double DEFAULT_ucrit = (1e4*ICM2M);
const double DEFAULT_neff  = 1.0;
/*--------------------------------------------------------------------------*/
MOS_COMMON::MOS_COMMON()
{
  lo	= NOT_INPUT;
  wo	= NOT_INPUT;
  ad_in	= NOT_INPUT;
  as_in	= NOT_INPUT;
  pd	= 0.0;
  ps	= 0.0;
  nrd	= 1.0;
  nrs	= 1.0;
  off = icset = false;
  for (int i = 0;  i < mNUM_INIT_COND;  i++)
    ic[i] = 0.;
}
/*--------------------------------------------------------------------------*/
MODEL_MOS::MODEL_MOS(const char *name):MODEL_DIODE(name)
{
  			/* inherit, then override diode defaults */
  cjo    = NOT_INPUT;	/* override 0.0 */
  pb     = 0.8;		/* override 1.0 */
  fcpb   = fc * pb;
			/* mos (not diode) defaults */
  vto    = NOT_INPUT;	/*calc*/
  kp     = NOT_INPUT;	/*calc*/
  gamma  = NOT_INPUT;	/*calc*/
  phi    = NOT_INPUT;	/*calc*/
  lambda = NOT_INPUT;
  rd     = NOT_INPUT;
  rs     = NOT_INPUT;
  cbd    = NOT_INPUT;
  cbs    = NOT_INPUT;
  is     = NOT_INPUT;
  cgso   = 0.0;
  cgdo   = 0.0;
  cgbo   = 0.0;
  rsh    = NOT_INPUT;
  js     = NOT_INPUT;
  tox    = DEFAULT_tox;
  nsub   = NOT_INPUT;	/*flag*/
  nss    = DEFAULT_nss;
  nfs    = 0.0;
  xj     = NOT_INPUT;
  ld     = 0.0;
  uo     = (600.*CM2M2);
  ucrit  = DEFAULT_ucrit;
  uexp   = NOT_INPUT;
  utra   = NOT_INPUT;
  vmax   = NOT_INPUT;
  neff   = DEFAULT_neff;
  kf     = NOT_INPUT;
  af     = NOT_INPUT;
  delta  = 0.0;
  theta  = 0.0;
  eta    = 0.0;
  kappa  = 0.2;
  level  = 2;
  tpg    = gSAME;
  polarity = pN;
  alpha	 = NOT_INPUT;
  xd     = NOT_INPUT;	/*calc if nsub spec, else not*/
  cox    = (E_OX / tox);
  vfb    = NOT_INPUT;
  vbi    = NOT_INPUT;
  xwb    = .25e-6;
  vbp    = NOT_INPUT;
  cfsox  = NOT_INPUT;
  delta3 = 0.0;
  sqrt_phi = NOT_INPUT;
  phi_sqrt_phi = NOT_INPUT;
  calc.vto = calc.kp = calc.gamma = calc.phi = calc.cj = false;
}
/*--------------------------------------------------------------------------*/
void MODEL_MOS::parse(CS& cmd)
{
  cmd.skiparg();	/* skip known ".model" */
  cmd.ctostr(label, LABELEN, TOKENTERM);
  cmd.stuck();
  /**/ set(cmd, "NMos", &polarity, pN)
    || set(cmd, "PMos", &polarity, pP);
  if (cmd.stuck()){
    cmd.check(bWARNING);
  }
  cmd.skiplparen();
  cmd.stuck();
  do{
    cmd.get("LEvel",  &level);
    cmd.get("VTO",    &vto,	mSCALE, (double)polarity);
    cmd.get("KP",     &kp);
    cmd.get("GAmma",  &gamma);
    cmd.get("PHI",    &phi,	mPOSITIVE);
    cmd.get("LAmbda", &lambda);
    cmd.get("RD",     &rd);
    cmd.get("RS",     &rs);
    cmd.get("CBD",    &cbd);
    cmd.get("CBS",    &cbs);
    cmd.get("IS",     &is);
    cmd.get("PB",     &pb,	mPOSITIVE);
    cmd.get("CGSo",   &cgso);
    cmd.get("CGDo",   &cgdo);
    cmd.get("CGBo",   &cgbo);
    cmd.get("RSH",    &rsh);
    cmd.get("CJ",     &cjo);
    cmd.get("MJ",     &mj);
    cmd.get("CJSw",   &cjsw);
    cmd.get("MJSw",   &mjsw);
    cmd.get("JS",     &js);
    cmd.get("TOX",    &tox,	mPOSITIVE);
    cmd.get("NSUb",   &nsub,	mSCALE, ICM2M3);
    cmd.get("NSS",    &nss,	mSCALE, ICM2M2);
    cmd.get("NFS",    &nfs,	mSCALE, ICM2M2);
    cmd.get("TPG",    (int*)&tpg);
    cmd.get("XJ",     &xj,	mPOSITIVE);
    cmd.get("LD",     &ld);
    cmd.get("UO",     &uo,	mSCALE, CM2M2);
    cmd.get("UCRit",  &ucrit,	mSCALE, ICM2M);
    cmd.get("UEXp",   &uexp);
    cmd.get("UTRa",   &utra);
    cmd.get("VMAx",   &vmax);
    cmd.get("NEFf",   &neff,	mPOSITIVE);
    cmd.get("KF",     &kf);
    cmd.get("AF",     &af);
    cmd.get("FC",     &fc);
    cmd.get("DELta",  &delta);
    cmd.get("THEta",  &theta);
    cmd.get("ETA",    &eta);
    cmd.get("KAPpa",  &kappa);
  }while (cmd.more() && !cmd.stuck());
  cmd.skiprparen();
  cmd.check(bWARNING);
  
  if ((rs == NOT_INPUT)  &&  (rd != NOT_INPUT)){
    error(bWARNING, "%s: rd input, but not rs. setting rs = 0.\n",
	  printlabel());
    rs = 0.;
  }else if ((rd == NOT_INPUT)  &&  (rs != NOT_INPUT)){
    error(bWARNING, "%s: rs input, but not rd. setting rd = 0.\n",
	  printlabel());
    rd = 0.;
  }
  
  if ((rsh != NOT_INPUT)  &&  (rd != NOT_INPUT)){
    error(bWARNING, "%s: rsh - rs - rd conflict: using %s\n",
	  printlabel(),
	  ((rd <= 0.)  &&  (rs <= 0.)) ? "rsh" : "rs,rd" );
  }else if ((rsh == NOT_INPUT)  &&  (rd == NOT_INPUT)){
    rsh = 0.;
  }
  
  if (tox == 0.){
    error(bWARNING, "%s: tox == 0: using %s\n",
	  printlabel(),
	  ftos(DEFAULT_tox, 0, 7, 0));
    tox = DEFAULT_tox;
  }
  cox = E_OX / tox;
  
  if (kp == NOT_INPUT){
    kp = uo * cox;
    calc.kp = true;
  }
  
  if (is == NOT_INPUT  &&  js == NOT_INPUT){
    is = mDEFAULT_is;
  }else if (is != NOT_INPUT  &&  js != NOT_INPUT){
    error(bWARNING, "%s: is - js conflict\n", printlabel());
  }
  
  if (nsub != NOT_INPUT){
    nsub = fabs(nsub);
    if (nsub < NI){
      error(bWARNING, "%s: nsub < ni\n", printlabel());
      nsub = NI;
    }
    if (cjo == NOT_INPUT){
      cjo = sqrt(E_SI * Q * nsub / (2. * pb ));
      calc.cj = true;
    }
    if (phi == NOT_INPUT){
      phi = (2.*K/Q)*OPT::tempamb*log(nsub/NI);
      if (phi < .1){
	error(bWARNING, "%s: calculated phi too small, using .1\n",
	      printlabel());
	phi = .1;
      }
      calc.phi = true;
    }
    if (gamma == NOT_INPUT){
      gamma = sqrt(2. * E_SI * Q * nsub) / cox;
      calc.gamma = true;
    }
    if (vto == NOT_INPUT){
      double egap = 
	1.16 - (7.02e-4*OPT::tempamb*OPT::tempamb) / (OPT::tempamb+1108.);
      double phi_ms = (tpg == gMETAL)
	? -.05 - (egap + polarity * phi) / 2.
	: -polarity * (tpg * egap + phi) / 2.;
      
      vfb = phi_ms - Q * nss / cox;
      vto = vfb + polarity * (phi + gamma * sqrt(phi));
      calc.vto = true;
    }
    alpha = (2. * E_SI / Q) / nsub;
    xd = sqrt(alpha);
    xwb = xd * sqrt(pb);
  }else{
    if (vto == NOT_INPUT)
      vto = DEFAULT_vto;
    if (gamma == NOT_INPUT)
      gamma = DEFAULT_gamma;
    if (phi == NOT_INPUT)
      phi = DEFAULT_phi;
    if (cjo == NOT_INPUT)
      cjo = DEFAULT_cj;
    alpha = 0.;
    xd = xwb / sqrt(pb);
  }
  sqrt_phi = sqrt(phi);
  phi_sqrt_phi = phi * sqrt_phi;
  if (lambda != NOT_INPUT  &&  lambda > .2)
    error(bWARNING, "%s: lambda too large (> .2)\n", printlabel());
  if (vfb == NOT_INPUT)
    vfb = vto - (phi + gamma * sqrt_phi);
  vbi = vfb + phi;
  vbp = ucrit * E_SI / cox;
  cfsox = Q * nfs / cox;
  fcpb = fc * pb;

  delta3 = delta * kPId2 * E_SI / cox;	/* level 3 stuff */
}
/*--------------------------------------------------------------------------*/
void MODEL_MOS::print(OMSTREAM where, int)const
{
  where.setfloatwidth(7);
  where   << ".model  " << label;
  where   << ((polarity < 0) ? "  pmos  (" : "  nmos  (");
  where   <<   "level=" << level;
  if (!calc.vto)
    where << "  vto="	<< vto*polarity;
  if (!calc.kp)
    where << "  kp="	<< kp;
  if (!calc.gamma)
    where << "  gamma="	<< gamma;
  if (!calc.phi)
    where << "  phi="	<< phi;
  if (lambda != NOT_INPUT)
    where << "  lambda="<< lambda;
  if (rd != NOT_INPUT)
    where << "  rd="	<< rd;
  if (rs != NOT_INPUT)
    where << "  rs="	<< rs;
  if (cbd != NOT_INPUT)
    where << "  cbd="	<< cbd;
  if (cbs != NOT_INPUT)
    where << "  cbs="	<< cbs;
  if (is != NOT_INPUT)
    where << "  is="	<< is;
  where   << "  pb="	<< pb;
  where   << "  cgso="	<< cgso;
  where   << "  cgdo="	<< cgdo;
  where   << "  cgbo="	<< cgbo;
  if (rsh != NOT_INPUT)
    where << "  rsh="	<< rsh;
  if (!calc.cj)
    where << "  cj="	<< cjo;
  where   << "  mj="	<< mj;
  where   << "  cjsw="	<< cjsw;
  where   << "  mjsw="	<< mjsw;
  if (js != NOT_INPUT)
    where << "  js="	<< js;
  where   << "  tox="	<< tox;
  if (nsub != NOT_INPUT)
    where << "  nsub="	<< nsub/ICM2M3;
  if (nss != DEFAULT_nss  ||  nsub != NOT_INPUT)
    where << "  nss="	<< nss /ICM2M2;
  where   << "  nfs="	<< nfs /ICM2M2;
  where   << "  tpg="   << tpg;
  if (xj != NOT_INPUT)
    where << "  xj="	<< xj;
  where   << "  ld="	<< ld;
  where   << "  uo="	<< uo   /CM2M2;
  if (ucrit != DEFAULT_ucrit  ||  uexp != NOT_INPUT)
    where << "  ucrit="	<< ucrit/ICM2M;
  if (uexp != NOT_INPUT)
    where << "  uexp="	<< uexp;
  if (utra != NOT_INPUT)
    where << "  utra="	<< utra;
  if (vmax != NOT_INPUT)
    where << "  vmax="	<< vmax;
  if (neff != DEFAULT_neff  ||  lambda == NOT_INPUT)
    where << "  neff="	<< neff;
  if (kf != NOT_INPUT)
    where << "  kf="	<< kf;
  if (af != NOT_INPUT)
    where << "  af="	<< af;
  where   << "  fc="	<< fc;
  where   << "  delta="	<< delta;
  if (level == 3){
    where << "  theta="	<< theta;
    where << "  eta="	<< eta;
    where << "  kappa="	<< kappa;
  }
  where   << ")\n*+(";
  where   << "* vfb="	<< vfb;
  if (calc.vto)
    where << "* vto="	<< vto*polarity;
  if (calc.kp)
    where << "* kp="	<< kp;
  if (calc.gamma)
    where << "* gamma="	<< gamma;
  if (calc.phi)
    where << "* phi="	<< phi;
  if (calc.cj)
    where << "* cj="	<< cjo;
  where << ")\n";
}
/*--------------------------------------------------------------------------*/
void MODEL_MOS::tr_eval(COMPONENT* brh)const
{
  DEV_MOS* d = (DEV_MOS*)brh;
  assert(d);
  switch (level){
    default:
      error(bDANGER, "%s: level %s not implemented, using 1\n",
	    printlabel(), level);
    case 1:	eval_mos1(d);	break;
    case 2:	eval_mos2(d);	break;
    case 3:	eval_mos3(d);	break;
  }
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
