/*$Id: l_bheap.h,v 12.1 1998/10/16 20:52:37 al Exp $ -*- C++ -*-
 * A binary heap
 */
#ifndef M_BHEAP_H
#define M_BHEAP_H
#include <assert.h>
#include "md_bool.h"
/*--------------------------------------------------------------------------*/
template <class T>
class BHEAP {
private:
  T  	  *q;           // storage array, allocated with "new"
  int     size;         // how big is it
  int	  rear;		// the first empty slot
public:
  	     BHEAP(int s=100)
	       {q = new T[s+1]; assert(q); rear = 1; size = s;}
	     ~BHEAP()		{delete [] q;}
  int        Count()const	{return rear-1;}
  int        Capacity()const	{return size;}
  int	     SpaceLeft()const	{return Capacity()-Count();}
  BHEAP<T>&  Clear()		{rear = 1; return *this;}
  bool       IsEmpty()const	{return Count()==0;}
  bool       IsFull()const	{return SpaceLeft()==0;}
  BHEAP<T>&  Insert(T x);
  BHEAP<T>&  operator<<(T x)	{return Insert(x);}
  T	     Dequeue();
  BHEAP<T>&  operator>>(T& x)	{x = Dequeue(); return *this;}
  BHEAP<T>&  DelFront()		{Dequeue(); return *this;}
  T	     Front()		{assert(!IsEmpty()); return q[1];}
};
/*--------------------------------------------------------------------------*/
template <class T>
inline BHEAP<T>& BHEAP<T>::Insert(T x)
{
  assert(!IsFull());
  int i;
  for (i = rear++; i > 1 && q[i/2] > x;  i /= 2){
    q[i] = q[i/2];
  }
  q[i] = x;
  return *this;
}
/*--------------------------------------------------------------------------*/
template <class T>
inline T BHEAP<T>::Dequeue()
{
  int i, child;
  assert(!IsEmpty());
  T min_element = q[1];
  T last_element = q[--rear];
  
  for (i=child=1;  i*2 <= rear;  i = child){
    child = i*2;
    if (child < rear){
      if (q[child+1] < q[child]){
        ++child;
      }
    }
    if (last_element > q[child]){
      q[i] = q[child];
    }else{
      break;
    }
  }
  q[i] = last_element;
  return min_element;
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
#endif
